module Cats
  module Core
    class Stack < ApplicationRecord
      # Stack statuses
      RESERVED = 'Reserved'.freeze
      ALLOCATED = 'Allocated'.freeze
      DESTROYED = 'Destroyed'.freeze
      STACK_STATUSES = [RESERVED, ALLOCATED, DESTROYED].freeze

      belongs_to :commodity
      belongs_to :store
      has_many :dispatch_transactions, foreign_key: :source_id
      has_many :receipt_transactions, foreign_key: :destination_id

      validates :code, :length, :width, :height, :start_x, :start_y, :commodity_status, :stack_status,
                :quantity, presence: true
      validates :code, uniqueness: true
      validates :length, :width, :height, :start_x, :start_y, numericality: { greater_than: 0 }
      validates :quantity, numericality: { greater_than_or_equal_to: 0 }
      validates :commodity_status, inclusion: { in: Cats::Core::Commodity::COMMODITY_STATUSES }
      validates :stack_status, inclusion: { in: STACK_STATUSES }
      validate :validate_coordinates, :validate_dimensions, :validate_distance_from_wall, :validate_overlap,
               :validate_space_between_stack

      def validate_coordinates
        return unless store.present? && length.present? && width.present? && start_x.present? && start_y.present?

        errors.add(:start_x, 'cannot exceed length') if start_x.present? && start_x > store.length
        errors.add(:start_y, 'cannot exceed width') if start_y.present? && start_y > store.width
      end

      def validate_dimensions
        return unless store.present? && length.present? && width.present? && start_x.present? && start_y.present?

        errors.add(:length, 'cannot exceed store length') if start_x + length > store.length
        errors.add(:width, 'cannot exceed store width') if start_y + width > store.width
      end

      def stacking_rules
        rule = Cats::Core::StackingRule.first

        # If no rule is defined, then define a default rule
        rule ||= Cats::Core::StackingRule.new(
          space_between_stack: 1,
          distance_from_gangway: 2,
          distance_from_ceiling: 1,
          maximum_height: 5,
          maximum_length: 5,
          maximum_width: 5,
          distance_from_wall: 1
        )
        rule
      end

      def validate_distance_from_wall
        return unless store.present? && length.present? && width.present? && start_x.present? && start_y.present?

        rule = stacking_rules
        if start_x < rule.distance_from_wall || store.length - (start_y + length) < rule.distance_from_wall ||
           store.width - (start_x + width) < rule.distance_from_wall || start_y < rule.distance_from_wall
          errors.add(:base,
                     message: "The stack must be placed #{rule.distance_from_wall} meter away from a store wall")
        end
      end

      def validate_overlap
        return unless store.present? && length.present? && width.present? && start_x.present? && start_y.present?

        stacks = store.stacks.where(stack_status: [RESERVED, ALLOCATED]).where.not(id: id)
        stacks.each do |s|
          errors.add(:base, message: "#{code} overlaps with #{s.code}") if overlaps?(self, s)
        end
      end

      def validate_space_between_stack
        return unless store.present? && length.present? && width.present? && start_x.present? && start_y.present?

        stacks = store.stacks.where(stack_status: [RESERVED, ALLOCATED]).where.not(id: id)
        rule = stacking_rules
        new_stack = dup
        new_stack.assign_attributes(start_x: start_x - rule.space_between_stack,
                                    start_y: start_y - rule.space_between_stack,
                                    length: length + (2 * rule.space_between_stack),
                                    width: width + (2 * rule.space_between_stack))
        stacks.each do |s|
          errors.add(:base, message: "#{code} overlaps with #{s.code}") if overlaps?(new_stack, s)
        end
      end

      # A method that checks if an overlap exists b/n two stacks.
      # An overlap b/n stacks does not occur if one of the stacks is completely above the other stack or
      # if one of the stacks is on the left side of the other stack.
      # If both conditions fail, it means an overlap exists b/n the two stacks

      def overlaps?(st1, st2)
        return false if st1.start_x > (st2.start_x + st2.width) || st2.start_x > (st1.start_x + st1.width)

        return false if st1.start_y > (st2.start_y + st2.length) || st2.start_y > (st1.start_y + st1.length)

        true
      end
    end
  end
end
